!------------------------------------------------------------------------------|
!------------------------------------------------------------------------------|
!                                                                              |
!              ||===\\                                                         | 
!              ||    \\                                                        |
!              ||     ||   //==\\   ||  ||   //==||  ||/==\\                   |
!              ||     ||  ||    ||  ||  ||  ||   ||  ||    ||                  |
!              ||    //   ||    ||  ||  ||  ||   ||  ||                        |
!              ||===//     \\==//    \\==\\  \\==\\  ||                        |
!                                                                              |
!------------------------------------------------------------------------------|
!------------------------------------------------------------------------------|
!                                                                              |
!              DEFINE_SURFACE    Nov. 2006                                     |
!                                                                              |
!------------------------------------------------------------------------------|
!------------------------------------------------------------------------------|

subroutine retriangulate_surface (surface,is,istep,params)

!------------------------------------------------------------------------------|
!(((((((((((((((( Purpose of the routine  ))))))))))))))))))))))))))))))))))))))
!------------------------------------------------------------------------------|
! If move_cloud in combination with regular remove_points failed (one
! node has too many neighbors), then this routine removes all points
! lying outside of the domain and recalculates the triangulation. 
! WARNING: This does not work for closed surfaces

!------------------------------------------------------------------------------|
!((((((((((((((((  declaration of the subroutine arguments  ))))))))))))))))))))
!------------------------------------------------------------------------------|

!use constants
use definitions
!use mpi

implicit none

include 'mpif.h'

type (sheet) :: surface
type (parameters) :: params
integer :: is, istep

!------------------------------------------------------------------------------|
!(((((((((((((((( declaration of the subroutine internal variables )))))))))))))
!------------------------------------------------------------------------------|

integer :: nproc,iproc,ierr
integer :: err,ntmax,nhmax,npmax,nmax,nmode,nvmax,nnpnmax,nh,nohalt_hull,loc
integer :: dmode,nt
integer :: nnn(1),nnlist(1),ntrilist(1)
integer,dimension(:),  allocatable :: hulltriangles,vis_tlist,vis_elist
integer,dimension(:),  allocatable :: add_tlist
integer,dimension(:,:),allocatable :: vertices,neighbour

double precision :: eps
double precision,dimension(:),  allocatable :: field
double precision,dimension(:,:),allocatable :: points,centres

logical clockwise
logical*1,dimension(:),allocatable :: lt_work,ln_work

!------------------------------------------------------------------------------|
!------------------------------------------------------------------------------|

call mpi_comm_size (mpi_comm_world,nproc,ierr)
call mpi_comm_rank (mpi_comm_world,iproc,ierr)

!if (params%debug>=2) call output_surf (surface,is,'faulty',istep,-1) 
if (params%debug>=1) call output_surf (surface,is,'faulty',istep,-1) 

!-------------------------------------------|
!-----building icon array-------------------|
!-------------------------------------------|
! we use the NN library to compute the triangulation 
! of points for the surface
!
! Code taken from create_surfaces.f90

if (iproc==0) write (8,'(a)') 'Starting triangulation'

ntmax=surface%nsurface*3
nhmax=surface%nsurface
npmax=surface%nsurface
nnpnmax=100
nmax=3*ntmax+npmax
nvmax=ntmax
allocate (points(2,surface%nsurface),stat=err) ; if (err.ne.0) call stop_run ('Error alloc points in retriangulate_surface')
allocate (field(surface%nsurface),stat=err)    ; if (err.ne.0) call stop_run ('Error alloc field in retriangulate_surface')
allocate (vertices(3,ntmax),stat=err)          ; if (err.ne.0) call stop_run ('Error alloc vertices in retriangulate_surface')
allocate (centres(3,ntmax),stat=err)           ; if (err.ne.0) call stop_run ('Error alloc centress in retriangulate_surface')
allocate (neighbour(3,ntmax),stat=err)         ; if (err.ne.0) call stop_run ('Error alloc neighbour in retriangulate_surface')
allocate (hulltriangles(nhmax),stat=err)       ; if (err.ne.0) call stop_run ('Error alloc hulltriangles in retriangulate_surface')
allocate (vis_tlist(nvmax),stat=err)           ; if (err.ne.0) call stop_run ('Error alloc vis_tlist in retriangulate_surface')
allocate (vis_elist(nvmax),stat=err)           ; if (err.ne.0) call stop_run ('Error alloc vis_elist in retriangulate_surface')
allocate (add_tlist(nvmax),stat=err)           ; if (err.ne.0) call stop_run ('Error alloc add_tlist in iretriangulate_surface')
allocate (lt_work(ntmax),stat=err)             ; if (err.ne.0) call stop_run ('Error alloc lt_work in retriangulate_surface')
allocate (ln_work(surface%nsurface),stat=err)  ; if (err.ne.0) call stop_run ('Error alloc ln_work in retriangulate_surface')
points(1,:)=surface%x
points(2,:)=surface%y
dmode=-2
nmode=0
loc=1
eps=1.0d-8
nohalt_hull=0
clockwise=.true.
call nn2d_setup (surface%nsurface,ntmax,nhmax,npmax,nnpnmax,nmax, &
                 points,dmode,nmode,clockwise,field,nt,vertices, &
                 centres,neighbour,nh,hulltriangles,nohalt_hull, &
                 loc,nnn,nnlist,ntrilist, &
                 eps,nvmax,vis_tlist,vis_elist,add_tlist, &
                 lt_work,ln_work)
if (iproc.eq.0) then
    write (*,'(a,i2,a,i8,a)') 'retriangulated surface ',is,' has ',nt,' triangles'
    write (8,'(a,i2,a,i8,a)') 'retriangulated surface ',is,' has ',nt,' triangles'
end if

if (surface%nt /= nt) then
    deallocate (surface%icon)
    allocate (surface%icon(3,nt),stat=err); if (err/=0) call stop_run ('Error alloc icon in retriangulate_surface')
    surface%nt = nt
end if

surface%icon(1:3,1:nt)=vertices(1:3,1:nt)

deallocate (points)
deallocate (field)
deallocate (vertices)
deallocate (centres)
deallocate (neighbour)
deallocate (hulltriangles)
deallocate (vis_tlist)
deallocate (vis_elist)
deallocate (add_tlist)
deallocate (lt_work)
deallocate (ln_work)

             
!----------------------------------------------|
!----computing the normals---------------------|  
!----------------------------------------------|

call compute_normals (surface%nsurface,surface%x,surface%y,surface%z, &
                      surface%nt,surface%icon,surface%xn,surface%yn,surface%zn)

!if (params%debug>=2) call output_surf (surface,is,'init',-1,-1) 
if (params%debug>=1) call output_surf (surface,is,'redone',istep,-1) 

end subroutine retriangulate_surface

